from abc import ABC, abstractmethod

import autograd.numpy as np
import autograd.numpy.random as npr
from autograd import value_and_grad, vector_jacobian_product, make_hvp
from autograd.core import getval

__all__ = [
    'VariationalObjective',
    'StochasticVariationalObjective',
    'ExclusiveKL',
    'DISInclusiveKL',
    'AlphaDivergence'
]


class VariationalObjective(ABC):
    """A class representing a variational objective to minimize"""

    def __init__(self, approx, model):
        """
        Parameters
        ----------
        approx : `ApproximationFamily` object
        model : `Model` object
        """
        self._approx = approx
        self._model = model
        self._objective_and_grad = None
        self._update_objective_and_grad()

    def __call__(self, var_param):
        """Evaluate objective and its gradient.

        May produce an (un)biased estimate of both.

        Parameters
        ----------
        var_param : `numpy.ndarray`, shape (var_param_dim,)
            The variational parameter.
        """
        if self._objective_and_grad is None:
            raise RuntimeError("no objective and gradient available")
        return self._objective_and_grad(var_param)

    @abstractmethod
    def _update_objective_and_grad(self):
        """Update the objective and gradient function.

        Should be called whenever a parameter that the objective depends on
        (e.g., `approx` or `model`) is updated."""
    
    def _hessian_vector_product(self, var_param, x):
        """Compute hessian vector product at given variaitonal parameter point x. """
        pass
    
    def update(self, var_param, direction):
        """Update the variational parameter in optimization."""
        return var_param - direction
        
    @property
    def approx(self):
        """The approximation family."""
        return self._approx

    @approx.setter
    def approx(self, value):
        self._approx = value
        self._update_objective_and_grad()

    @property
    def model(self):
        """The model."""
        return self._model

    @model.setter
    def model(self, value):
        self._model = value
        self._update_objective_and_grad()


class StochasticVariationalObjective(VariationalObjective):
    """A class representing a variational objective approximated using Monte Carlo."""

    def __init__(self, approx, model, num_mc_samples):
        """
        Parameters
        ----------
        approx : `ApproximationFamily` object
        model : `Model` object
        num_mc_sample : `int`
            Number of Monte Carlo samples to use to approximate the objective.
        """
        self._num_mc_samples = num_mc_samples
        super().__init__(approx, model)

    @property
    def num_mc_samples(self):
        """Number of Monte Carlo samples to use to approximate the objective."""
        return self._num_mc_samples

    @num_mc_samples.setter
    def num_mc_samples(self, value):
        self._num_mc_samples = value
        self._update_objective_and_grad()


class ExclusiveKL(StochasticVariationalObjective):
    """Exclusive Kullback-Leibler divergence.

    Equivalent to using the canonical evidence lower bound (ELBO)
    """

    def __init__(self, approx, model, num_mc_samples, use_path_deriv=False):
        """
        Parameters
        ----------
        approx : `ApproximationFamily` object
        model : `Model` object
        num_mc_sample : `int`
            Number of Monte Carlo samples to use to approximate the objective.
        use_path_deriv : `bool`
            Use path derivative (for "sticking the landing") gradient estimator
        """
        self._use_path_deriv = use_path_deriv
        super().__init__(approx, model, num_mc_samples)

    def _update_objective_and_grad(self):
        approx = self.approx

        def variational_objective(var_param):
            samples = approx.sample(var_param, self.num_mc_samples)
            if self._use_path_deriv:
                var_param_stopped = getval(var_param)
                lower_bound = np.mean(
                    self.model(samples) - approx.log_density(var_param_stopped, samples))
            elif approx.supports_entropy:
                lower_bound = np.mean(self.model(samples)) + approx.entropy(var_param)
            else:
                lower_bound = np.mean(self.model(samples) - approx.log_density(samples))
            return -lower_bound
        self._hvp = make_hvp(variational_objective)
        self._objective_and_grad = value_and_grad(variational_objective)

    def _hessian_vector_product(self, var_param, x):
        hvp_fun = self._hvp(var_param)[0]
        return hvp_fun(x)


class DISInclusiveKL(StochasticVariationalObjective):
    """Inclusive Kullback-Leibler divergence using Distilled Importance Sampling."""

    def __init__(self, approx, model, num_mc_samples, ess_target,
                 temper_prior, temper_prior_params, use_resampling=True,
                 num_resampling_batches=1, w_clip_threshold=10):
        """
        Parameters
        ----------
        approx : `ApproximationFamily` object
        model : `Model` object
        num_mc_sample : `int`
            Number of Monte Carlo samples to use to approximate the KL divergence.
            (N in the paper)
        ess_target: `int`
            The ess target to adjust epsilon (M in the paper). It is also the number of
            samples in resampling.
        temper_prior: `Model` object
            A prior distribution to temper the model. Typically multivariate normal.
        temper_prior_params: `numpy.ndarray` object
            Parameters for the temper prior. Typically mean 0 and variance 1.
        use_resampling: `bool`
            Whether to use resampling.
        num_resampling_batches: `int`
            Number of resampling batches. The resampling batch is `max(1, ess_target / num_resampling_batches)`.
        w_clip_threshold: `float`
            The maximum weight.
        """
        self._ess_target = ess_target
        self._w_clip_threshold = w_clip_threshold
        self._max_bisection_its = 50
        self._max_eps = self._eps = 1
        self._use_resampling = use_resampling
        self._num_resampling_batches = num_resampling_batches
        self._resampling_batch_size = max(1, self._ess_target // num_resampling_batches)
        self._objective_step = 0

        self._tempered_model_log_pdf = lambda eps, samples, log_p_unnormalized: (
            eps * temper_prior.log_density(temper_prior_params, samples)
            + (1 - eps) * log_p_unnormalized)
        super().__init__(approx, model, num_mc_samples)

    def _get_weights(self, eps, samples, log_p_unnormalized, log_q):
        """Calculates normalised importance sampling weights"""
        logw = self._tempered_model_log_pdf(eps, samples, log_p_unnormalized) - log_q
        max_logw = np.max(logw)
        if max_logw == -np.inf:
            raise ValueError('All weights zero! '
                             + 'Suggests overflow in importance density.')

        w = np.exp(logw)
        return w

    def _get_ess(self, w):
        """Calculates effective sample size of normalised importance sampling weights"""
        ess = (np.sum(w) ** 2.0) / np.sum(w ** 2.0)
        return ess

    def _get_eps_and_weights(self, eps_guess, samples, log_p_unnormalized, log_q):
        """Find new epsilon value

        Uses bisection to find epsilon < eps_guess giving required ESS.
        If none exists, returns eps_guess.

        Returns new epsilon value and corresponding ESS and normalised importance sampling weights.
        """

        lower = 0.
        upper = eps_guess
        eps_guess = (lower + upper) / 2.
        for i in range(self._max_bisection_its):
            w = self._get_weights(eps_guess, samples, log_p_unnormalized, log_q)
            ess = self._get_ess(w)
            if ess > self._ess_target:
                upper = eps_guess
            else:
                lower = eps_guess
            eps_guess = (lower + upper) / 2.

        w = self._get_weights(eps_guess, samples, log_p_unnormalized, log_q)
        ess = self._get_ess(w)

        # Consider returning extreme epsilon values if they are still endpoints
        if lower == 0.:
            eps_guess = 0.
        if upper == self._max_eps:
            eps_guess = self._max_eps

        return eps_guess, ess, w

    def _clip_weights(self, w):
        """Clip weights to `self._w_clip_threshold`
        Other weights are scaled up proportionately to keep sum equal to 1"""
        S = np.sum(w)
        if not any(w > S * self._w_clip_threshold):
            return w

        to_clip = (w >= S * self._w_clip_threshold)  # nb clip those equal to max_weight
        # so we don't push them over it!
        n_to_clip = np.sum(to_clip)
        to_not_clip = np.logical_not(to_clip)
        sum_unclipped = np.sum(w[to_not_clip])
        if sum_unclipped == 0:
            # Impossible to clip further!
            return w
        w[to_clip] = self._w_clip_threshold * sum_unclipped \
            / (1. - self._w_clip_threshold * n_to_clip)
        return self._clip_weights(w)

    def _update_objective_and_grad(self):
        approx = self.approx

        def variational_objective(var_param):
            if not self._use_resampling or self._objective_step % self._num_resampling_batches == 0:
                self._state_samples = getval(approx.sample(var_param, self.num_mc_samples))
                self._state_log_q = approx.log_density(var_param, self._state_samples)
                self._state_log_p_unnormalized = self.model(self._state_samples)

                self._eps, ess, w = self._get_eps_and_weights(
                    self._eps, self._state_samples, self._state_log_p_unnormalized, self._state_log_q)
                self._state_w_clipped = self._clip_weights(w)
                self._state_w_sum = np.sum(self._state_w_clipped)
                self._state_w_normalized = self._state_w_clipped / self._state_w_sum

            self._objective_step += 1

            if not self._use_resampling:
                return -np.inner(getval(self._state_w_clipped), self._state_log_q) / self.num_mc_samples
            else:
                indices = np.random.choice(self.num_mc_samples,
                                           size=self._resampling_batch_size, p=getval(self._state_w_normalized))
                samples_resampled = self._state_samples[indices]

                obj = np.mean(-approx.log_density(var_param, getval(samples_resampled)))

                return obj * getval(self._state_w_sum) / self.num_mc_samples

        self._objective_and_grad = value_and_grad(variational_objective)


class AlphaDivergence(StochasticVariationalObjective):
    """Log of the alpha-divergence."""

    def __init__(self, approx, model, num_mc_samples, alpha):
        """
        Parameters
        ----------
        approx : `ApproximationFamily` object
        model : `Model` object
        num_mc_sample : `int`
            Number of Monte Carlo samples to use to approximate the objective.
        alpha : `float`
        """
        self._alpha = alpha
        super().__init__(approx, model, num_mc_samples)

    @property
    def alpha(self):
        """Alpha parameter of the divergence."""
        return self._alpha

    def _update_objective_and_grad(self):
        """Provides a stochastic estimate of the variational lower bound."""
        def compute_log_weights(var_param, seed):
            samples = self.approx.sample(var_param, self.num_mc_samples, seed)
            log_weights = self.model(samples) - self.approx.log_density(var_param, samples)
            return log_weights

        log_weights_vjp = vector_jacobian_product(compute_log_weights)
        alpha = self.alpha
        # manually compute objective and gradient

        def objective_grad_and_log_norm(var_param):
            # must create a shared seed!
            seed = npr.randint(2**32)
            log_weights = compute_log_weights(var_param, seed)
            log_norm = np.max(log_weights)
            scaled_values = np.exp(log_weights - log_norm)**alpha
            obj_value = np.log(np.mean(scaled_values)) / alpha + log_norm
            obj_grad = alpha * log_weights_vjp(var_param, seed, scaled_values) / scaled_values.size
            return (obj_value, obj_grad)

        self._objective_and_grad = objective_grad_and_log_norm



"""Pareto smoothed importance sampling (PSIS)

This module implements Pareto smoothed importance sampling (PSIS) and PSIS
leave-one-out (LOO) cross-validation for Python (Numpy).

Included functions
------------------
psisloo
    Pareto smoothed importance sampling leave-one-out log predictive densities.

psislw
    Pareto smoothed importance sampling.

gpdfitnew
    Estimate the paramaters for the Generalized Pareto Distribution (GPD).

gpinv
    Inverse Generalised Pareto distribution function.

sumlogs
    Sum of vector where numbers are represented by their logarithms.

References
----------
Aki Vehtari, Andrew Gelman and Jonah Gabry (2017). Practical
Bayesian model evaluation using leave-one-out cross-validation
and WAIC. Statistics and Computing, 27(5):1413–1432.
doi:10.1007/s11222-016-9696-4. https://arxiv.org/abs/1507.04544

Aki Vehtari, Andrew Gelman and Jonah Gabry (2017). Pareto
smoothed importance sampling. https://arxiv.org/abs/arXiv:1507.02646v5

"""

from __future__ import division  # For Python 2 compatibility

import numpy as np

# 3-Clause BSD License
"""
Copyright 2017 Aki Vehtari, Tuomas Sivula

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors
may be used to endorse or promote products derived from this software without
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. """


def psisloo(log_lik, **kwargs):
    r"""PSIS leave-one-out log predictive densities.

    Computes the log predictive densities given posterior samples of the log
    likelihood terms :math:`p(y_i|\theta^s)` in input parameter `log_lik`.
    Returns a sum of the leave-one-out log predictive densities `loo`,
    individual leave-one-out log predictive density terms `loos` and an estimate
    of Pareto tail indeces `ks`. The estimates are unreliable if tail index
    ``k > 0.7`` (see more in the references listed in the module docstring).

    Additional keyword arguments are passed to the :meth:`psislw()` function
    (see the corresponding documentation).

    Parameters
    ----------
    log_lik : ndarray
        Array of size n x m containing n posterior samples of the log likelihood
        terms :math:`p(y_i|\theta^s)`.

    Returns
    -------
    loo : scalar
        sum of the leave-one-out log predictive densities

    loos : ndarray
        individual leave-one-out log predictive density terms

    ks : ndarray
        estimated Pareto tail indeces

    """
    # ensure overwrite flag in passed arguments
    kwargs['overwrite_lw'] = True
    # log raw weights from log_lik
    lw = -log_lik
    # compute Pareto smoothed log weights given raw log weights
    lw, ks = psislw(lw, **kwargs)
    # compute
    lw += log_lik
    loos = sumlogs(lw, axis=0)
    loo = loos.sum()
    return loo, loos, ks


def psislw(lw, Reff=1.0, overwrite_lw=False):
    """Pareto smoothed importance sampling (PSIS).

    Parameters
    ----------
    lw : ndarray
        Array of size n x m containing m sets of n log weights. It is also
        possible to provide one dimensional array of length n.

    Reff : scalar, optional
        relative MCMC efficiency ``N_eff / N``

    overwrite_lw : bool, optional
        If True, the input array `lw` is smoothed in-place, assuming the array
        is F-contiguous. By default, a new array is allocated.

    Returns
    -------
    lw_out : ndarray
        smoothed log weights
    kss : ndarray
        Pareto tail indices

    """
    if lw.ndim == 2:
        n, m = lw.shape
    elif lw.ndim == 1:
        n = len(lw)
        m = 1
    else:
        raise ValueError("Argument `lw` must be 1 or 2 dimensional.")
    if n <= 1:
        raise ValueError("More than one log-weight needed.")

    if overwrite_lw and lw.flags.f_contiguous:
        # in-place operation
        lw_out = lw
    else:
        # allocate new array for output
        lw_out = np.copy(lw, order='F')

    # allocate output array for kss
    kss = np.empty(m)

    # precalculate constants
    cutoff_ind = - int(np.ceil(min(0.2 * n, 3 * np.sqrt(n / Reff)))) - 1
    cutoffmin = np.log(np.finfo(float).tiny)
    np.log(n)
    k_min = 1 / 3

    # loop over sets of log weights
    for i, x in enumerate(lw_out.T if lw_out.ndim == 2 else lw_out[None, :]):
        # improve numerical accuracy
        x -= np.max(x)
        # sort the array
        x_sort_ind = np.argsort(x)
        # divide log weights into body and right tail
        xcutoff = max(
            x[x_sort_ind[cutoff_ind]],
            cutoffmin
        )
        expxcutoff = np.exp(xcutoff)
        tailinds, = np.where(x > xcutoff)
        x2 = x[tailinds]
        n2 = len(x2)
        if n2 <= 4:
            # not enough tail samples for gpdfitnew
            k = np.inf
        else:
            # order of tail samples
            x2si = np.argsort(x2)
            # fit generalized Pareto distribution to the right tail samples
            np.exp(x2, out=x2)
            x2 -= expxcutoff
            k, sigma = gpdfitnew(x2, sort=x2si)
        if k >= k_min and not np.isinf(k):
            # no smoothing if short tail or GPD fit failed
            # compute ordered statistic for the fit
            sti = np.arange(0.5, n2)
            sti /= n2
            qq = gpinv(sti, k, sigma)
            qq += expxcutoff
            np.log(qq, out=qq)
            # place the smoothed tail into the output array
            x[tailinds[x2si]] = qq
            # truncate smoothed values to the largest raw weight 0
            x[x > 0] = 0
        # renormalize weights
        x -= sumlogs(x)
        # store tail index k
        kss[i] = k

    # If the provided input array is one dimensional, return kss as scalar.
    if lw_out.ndim == 1:
        kss = kss[0]

    return lw_out, kss


def gpdfitnew(x, sort=True, sort_in_place=False, return_quadrature=False):
    """Estimate the paramaters for the Generalized Pareto Distribution (GPD)

    Returns empirical Bayes estimate for the parameters of the two-parameter
    generalized Parato distribution given the data.

    Parameters
    ----------
    x : ndarray
        One dimensional data array

    sort : bool or ndarray, optional
        If known in advance, one can provide an array of indices that would
        sort the input array `x`. If the input array is already sorted, provide
        False. If True (default behaviour), the array is sorted internally.

    sort_in_place : bool, optional
        If `sort` is True and `sort_in_place` is True, the array is sorted
        in-place (False by default).

    return_quadrature : bool, optional
        If True, quadrature points and weight `ks` and `w` of the marginal posterior distribution
        of k are also calculated and returned. False by default.

    Returns
    -------
    k, sigma : float
        estimated parameter values

    ks, w : ndarray
        Quadrature points and weights of the marginal posterior distribution
        of `k`. Returned only if `return_quadrature` is True.

    Notes
    -----
    This function returns a negative of Zhang and Stephens's k, because it is
    more common parameterisation.

    """
    if x.ndim != 1 or len(x) <= 1:
        raise ValueError("Invalid input array.")

    # check if x should be sorted
    if sort is True:
        if sort_in_place:
            x.sort()
            xsorted = True
        else:
            sort = np.argsort(x)
            xsorted = False
    elif sort is False:
        xsorted = True
    else:
        xsorted = False

    n = len(x)
    PRIOR = 3
    m = 30 + int(np.sqrt(n))

    bs = np.arange(1, m + 1, dtype=float)
    bs -= 0.5
    np.divide(m, bs, out=bs)
    np.sqrt(bs, out=bs)
    np.subtract(1, bs, out=bs)
    if xsorted:
        bs /= PRIOR * x[int(n / 4 + 0.5) - 1]
        bs += 1 / x[-1]
    else:
        bs /= PRIOR * x[sort[int(n / 4 + 0.5) - 1]]
        bs += 1 / x[sort[-1]]

    ks = np.negative(bs)
    temp = ks[:, None] * x
    np.log1p(temp, out=temp)
    np.mean(temp, axis=1, out=ks)

    L = bs / ks
    np.negative(L, out=L)
    np.log(L, out=L)
    L -= ks
    L -= 1
    L *= n

    temp = L - L[:, None]
    np.exp(temp, out=temp)
    w = np.sum(temp, axis=1)
    np.divide(1, w, out=w)

    # remove negligible weights
    dii = w >= 10 * np.finfo(float).eps
    if not np.all(dii):
        w = w[dii]
        bs = bs[dii]
    # normalise w
    w /= w.sum()

    # posterior mean for b
    b = np.sum(bs * w)
    # Estimate for k, note that we return a negative of Zhang and
    # Stephens's k, because it is more common parameterisation.
    temp = (-b) * x
    np.log1p(temp, out=temp)
    k = np.mean(temp)
    if return_quadrature:
        np.negative(x, out=temp)
        temp = bs[:, None] * temp
        np.log1p(temp, out=temp)
        ks = np.mean(temp, axis=1)
    # estimate for sigma
    sigma = -k / b * n / (n - 0)
    # weakly informative prior for k
    a = 10
    k = k * n / (n + a) + a * 0.5 / (n + a)
    if return_quadrature:
        ks *= n / (n + a)
        ks += a * 0.5 / (n + a)

    if return_quadrature:
        return k, sigma, ks, w
    else:
        return k, sigma


def gpinv(p, k, sigma):
    """Inverse Generalised Pareto distribution function."""
    x = np.empty(p.shape)
    x.fill(np.nan)
    if sigma <= 0:
        return x
    ok = (p > 0) & (p < 1)
    if np.all(ok):
        if np.abs(k) < np.finfo(float).eps:
            np.negative(p, out=x)
            np.log1p(x, out=x)
            np.negative(x, out=x)
        else:
            np.negative(p, out=x)
            np.log1p(x, out=x)
            x *= -k
            np.expm1(x, out=x)
            x /= k
        x *= sigma
    else:
        if np.abs(k) < np.finfo(float).eps:
            # x[ok] = - np.log1p(-p[ok])
            temp = p[ok]
            np.negative(temp, out=temp)
            np.log1p(temp, out=temp)
            np.negative(temp, out=temp)
            x[ok] = temp
        else:
            # x[ok] = np.expm1(-k * np.log1p(-p[ok])) / k
            temp = p[ok]
            np.negative(temp, out=temp)
            np.log1p(temp, out=temp)
            temp *= -k
            np.expm1(temp, out=temp)
            temp /= k
            x[ok] = temp
        x *= sigma
        x[p == 0] = 0
        if k >= 0:
            x[p == 1] = np.inf
        else:
            x[p == 1] = -sigma / k
    return x


def sumlogs(x, axis=None, out=None):
    """Sum of vector where numbers are represented by their logarithms.

    Calculates ``np.log(np.sum(np.exp(x), axis=axis))`` in such a fashion that
    it works even when elements have large magnitude.

    """
    maxx = x.max(axis=axis, keepdims=True)
    xnorm = x - maxx
    np.exp(xnorm, out=xnorm)
    out = np.sum(xnorm, axis=axis, out=out)
    if isinstance(out, np.ndarray):
        np.log(out, out=out)
    else:
        out = np.log(out)
    out += np.squeeze(maxx)
    return out



import warnings

import autograd.numpy as np
from scipy.fftpack import next_fast_len


def autocov(samples, axis=-1):
    """Compute autocovariance estimates for every lag for the input array.
    Parameters
    ----------
    samples : `numpy.ndarray(n_chains, n_iters)`
        An array containing samples
    Returns
    -------
    acov: `numpy.ndarray`
        Autocovariance of samples that has same size as the input array
    """
    axis = axis if axis > 0 else len(samples.shape) + axis
    n = samples.shape[axis]
    m = next_fast_len(2 * n)

    samples = samples - samples.mean(axis, keepdims=True)

    # added to silence tuple warning for a submodule
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        ifft_samp = np.fft.rfft(samples, n=m, axis=axis)
        ifft_samp *= np.conjugate(ifft_samp)

        shape = tuple(
            slice(None) if dim_len != axis else slice(0, n)
            for dim_len, _ in enumerate(samples.shape)
        )
        cov = np.fft.irfft(ifft_samp, n=m, axis=axis)[shape]
        cov /= n
        return cov


def ess(samples):
    """
    Computing Effective Sample Size

    Parameters
    ----------
     samples : `numpy.ndarray(n_chains, n_iters)`
        An array containing samples

    Returns
    -------
    ess : `real`
       effective sample size of the given sample

    """
    n_chain, n_draw = samples.shape
    if (n_chain > 1):
        ValueError("Number of chains must be 1")
    acov = autocov(samples, axis=1)
    # chain_mean = samples.mean(axis=1)
    mean_var = np.mean(acov[:, 0]) * n_draw / (n_draw - 1.0)
    var_plus = mean_var * (n_draw - 1.0) / n_draw
    # if n_chain > 1:
    #   var_plus += _numba_var(svar, np.var, chain_mean, axis=None, ddof=1)

    rho_hat_t = np.zeros(n_draw)
    rho_hat_even = 1.0
    rho_hat_t[0] = rho_hat_even
    rho_hat_odd = 1.0 - (mean_var - np.mean(acov[:, 1])) / var_plus
    rho_hat_t[1] = rho_hat_odd

    # Geyer's initial positive sequence
    t = 1
    while t < (n_draw - 3) and (rho_hat_even + rho_hat_odd) > 0.0:
        rho_hat_even = 1.0 - (mean_var - np.mean(acov[:, t + 1])) / var_plus
        rho_hat_odd = 1.0 - (mean_var - np.mean(acov[:, t + 2])) / var_plus
        if (rho_hat_even + rho_hat_odd) >= 0:
            rho_hat_t[t + 1] = rho_hat_even
            rho_hat_t[t + 2] = rho_hat_odd
        t += 2

    max_t = t - 2
    # improve estimation
    if rho_hat_even > 0:
        rho_hat_t[max_t + 1] = rho_hat_even
    # Geyer's initial monotone sequence
    t = 1
    while t <= max_t - 2:
        if (rho_hat_t[t + 1] + rho_hat_t[t + 2]) > (rho_hat_t[t - 1] + rho_hat_t[t]):
            rho_hat_t[t + 1] = (rho_hat_t[t - 1] + rho_hat_t[t]) / 2.0
            rho_hat_t[t + 2] = rho_hat_t[t + 1]
        t += 2

    ess = n_chain * n_draw
    tau_hat = -1.0 + 2.0 * np.sum(rho_hat_t[: max_t + 1]) + np.sum(rho_hat_t[max_t + 1: max_t + 2])
    tau_hat = max(tau_hat, 1 / np.log10(ess))
    ess = ess / tau_hat
    if np.isnan(rho_hat_t).any():
        ess = np.nan
    return ess


def MCSE(sample):
    """
    Compute the Monte Carlo standard error (MCSE)

    Parameters
    ----------
    samples : `numpy.ndarray(n_iters, 2*dim)`
        An array containing variational samples

    Returns
    -------
    mcse : `numpy.ndarray(2*dim)`
        MCSE for each variational parameter

    """
    n_iters, d = sample.shape
    sd_dev = np.sqrt(np.var(sample, ddof=1, axis=0))
    eff_samp = [ess(sample[:, i].reshape(1, n_iters)) for i in range(d)]
    mcse = sd_dev / np.sqrt(eff_samp)
    return eff_samp, mcse


def compute_R_hat(chains, warmup=0, jitter=1e-8):
    """
    Computing R hat values using split R hat approach

    Parameters
    ----------
    chains : `numpy_ndarray(n_iters, dimensions)`
        Sample of parameter estimates that has one chain
    warmup : `int`, optional
        Number of iterations needed for warmup. The default is 0.
    jitter : `float`, optional
        Smoothing term that avoids division by zero. The default is 1e-8.

    Returns
    -------
    R_hat : `numpy_ndarray(dimenstions,)`
        Computed R hat values for each parameter

    """
    n_chains = 1
    chains = chains[warmup:, :]
    n_iters, d = chains.shape
    if n_iters % 2 == 1:
        n_iters = int(n_iters - 1)
        chains = chains[:n_iters, :]
    n_iters = n_iters // 2
    n_chains2 = n_chains * 2
    psi = np.reshape(chains, (n_chains2, n_iters, d))
    psi_dot_j = np.mean(psi, axis=1)
    psi_dot_dot = np.mean(psi_dot_j, axis=0)
    s_j_2 = np.sum((psi - np.expand_dims(psi_dot_j, axis=1)) ** 2, axis=1) / (n_iters - 1)
    B = n_iters * np.sum((psi_dot_j - psi_dot_dot) ** 2, axis=0) / (n_chains2 - 1)
    W = np.nanmean(s_j_2, axis=0)
    W = W + jitter
    var_hat = (n_iters - 1) / n_iters + (B / (n_iters * W))
    R_hat = np.sqrt(var_hat)
    return R_hat


def R_hat_convergence_check(samples, windows, Rhat_threshold=1.1):
    """
    Convergence check of samples using R_hat

    Parameters
    ----------
    samples : `list`
        Samples to check the convergence
    windows : `numpy_array`
        Windows that should be used to do the check

    Returns
    -------
    success: `bool`
        Whether the convergence criterion is met
    best_W: `int`
        Best window size
    """
    R_hat_array = [np.max(compute_R_hat(np.array(samples[-window:]), 0)) for window in windows]
    best_R_hat_ind = np.argmin(R_hat_array)
    success = R_hat_array[best_R_hat_ind] <= Rhat_threshold
    return success, windows[best_R_hat_ind]



from autograd.extend import defvjp, primitive

from ._utils import ensure_2d, vectorize_if_needed

__all__ = [
    'Model',
    'StanModel'
]


class Model(object):
    """Base class for representing a model.

    Does not support tempering. It can be overridden in part or in whole by
    classes that inherit it. See ``StanModel`` for an example."""

    def __init__(self, log_density):
        """
        Parameters
        ----------
        log_density : `function`
            Function for computing the (unnormalized) log density of the model.
            Must support automatic differentiation with ``autograd``.
        """
        self._log_density = log_density

    def __call__(self, model_param):
        """Compute (unnormalized) log density of the model.

        Parameters
        ----------
        model_param : `numpy.ndarray`, shape (dim,)
            Model parameter value

        Returns
        -------
        log_density : `float`
        """
        return self._log_density(model_param)

    def constrain(self, model_param):
        """Construct dictionary of constrained parameters.

        Parameters
        ----------
        model_param : `numpy.ndarray`, shape (dim,)
            Model parameter value

        Returns
        -------
        constrained_params : `dict`

        Raises
        ------
        NotImplementedError
            If constrained parameterization is not supported.
        """
        raise NotImplementedError()

    @property
    def supports_tempering(self):
        """Whether the model supports tempering."""
        return False

    def set_inverse_temperature(self, inverse_temp):
        """If tempering supported, set inverse temperature.

        Parameters
        ----------
        inverse_temp : `float`

        Raises
        ------
        NotImplementedError
            If tempering is not supported.
        """
        raise NotImplementedError()


def _make_stan_log_density(fitobj):
    @primitive
    def log_density(x):
        return vectorize_if_needed(fitobj.log_prob, x)

    def log_density_vjp(ans, x):
        return lambda g: ensure_2d(g) * vectorize_if_needed(fitobj.grad_log_prob, x)
    defvjp(log_density, log_density_vjp)
    return log_density


class StanModel(Model):
    """Class that encapsulates a PyStan model."""

    def __init__(self, fit):
        """
        Parameters
        ----------
        fit : `StanFit4model` object
        """
        self._fit = fit
        super().__init__(_make_stan_log_density(fit))

    def constrain(self, model_param):
        return self._fit.constrain_pars(model_param)



from warnings import warn

import numpy as np

__all__ = [
    'all_diagnostics',
    'error_bounds',
    'wasserstein_bounds',
    'divergence_bound'
]


def all_diagnostics(log_weights, *, samples=None, moment_bound_fn=None,
                    q_var=None, p_var=None, log_norm_bound=None):
    """Compute all VI diagnostics.

    Compute error and distance bounds between distribution `p` and `q` using
    samples from `q`. The distributions need not be normalized.
    Also compute the Pareto k-hat diagnostic.

    Parameters
    ----------
    log_weights : `array-like` of `int`, shape (n_samples,)
        log weights `log p(x_i) - log q(x_i)`, where `x_i` is sampled from `q`
        and `p` may be an unnormalized distribution

    samples : array-like matrix, shape (n_samples, n_dimensions)
        samples `x_i` associated with log weights

    moment_bound_fn : `function`
        `moment_bound_fn(p)` should return a bound on
        :math:`\\min_y \\mathbb E[(x_i - y)^p]`.
        It must be provided if `samples` is `None` and it must support `p = 2`
        and `p = 4`.

    q_var : `float` or `array-like matrix`
        (Bound on) the (co)variance of `q`.

    p_var : `float` or `array-like matrix`
        (Bound on) the (co)variance of `p`.

    log_norm_bound : `float`
        Bound on the overall log normalization constant (the log marginal
        likelihood when `p` is the unnormalized log posterior)


    Returns
    -------
    results : `dict`
        contains the following entries: `mean_error`, `var_error`, `std_error`,
        `d2`, `W1`, `W2`."""
    d2, log_norm_bound = divergence_bound(log_weights,
                                          log_norm_bound=log_norm_bound,
                                          return_log_norm_bound=True)
    results = wasserstein_bounds(d2, samples=samples,
                                 moment_bound_fn=moment_bound_fn)

    if q_var is None and samples is not None:
        q_var = np.cov(samples.T)

    results.update(error_bounds(q_var=q_var, p_var=p_var, **results))
    results['d2'] = d2
    results['log_norm_bound'] = log_norm_bound
    return results


def _compute_norm_if_needed(var):
    if np.asarray(var).ndim == 2:
        return np.linalg.norm(var, ord=2)
    return var


def error_bounds(*, W1=np.inf, W2=np.inf, q_var=np.inf, p_var=np.inf):
    """Compute error bounds.

    Compute bounds on differences in the means, standard deviations, and
    covariances of `p` and `q` using (bounds on) the 1- and 2-Wasserstein
    distance.

    Parameters
    ----------
    W1 : `float`
        (Bound on) the 1-Wasserstein distance between `p` and `q`.

    W2 : `float`
        (Bound on) the 2-Wasserstein distance between `p` and `q`.

    q_var : `float` or `array-like matrix`
        (Bound on) the (co)variance of `q`.

    p_var : `float` or `array-like matrix`
        (Bound on) the (co)variance of `p`.

    Returns
    -------
    results : `dict`
        contains the following bounds: `mean_error`, `var_error`, `std_error`."""
    results = dict()
    results['mean_error'] = mean_bound(min(W1, W2))
    results['std_error'] = std_bound(W2)
    results['cov_error'] = var_bound(W2, _compute_norm_if_needed(q_var),
                                     _compute_norm_if_needed(p_var))
    return results


def wasserstein_bounds(d2, *, samples=None, moment_bound_fn=None):
    """Compute all bounds.

    Compute 1- and 2-Wasserstein distance bounds between distribution `p` and
    `q` using a bound on the 2-divergence and moment bounds.

    Parameters
    ----------
    d2 : `float`
        (Bound on) the 2-divergence between `p` and `q`.

    samples : `array-like matrix`, shape (n_samples, n_dimensions)
        samples from `q`.

    moment_bound_fn : `array-like matrix`, shape (n_variant_types, n_signatures)
        `moment_bound_fn(a)` should return a bound on
        :math:`\\min_y \\mathbb E[(x_i - y)^p]`.
        It must be provided if `samples` is `None`. Must support `a = 2`
        and `a = 4`.

    Returns
    -------
    results : `dict`
        contains the following bounds: `W1`, `W2`."""
    results = dict()
    if moment_bound_fn is None:
        if samples is None:
            raise ValueError('must provides samples if moment_bound_fn not given')
        samples = np.asarray(samples)
        if samples.ndim == 1:
            samples = samples[:, np.newaxis]
        sample_mean = np.mean(samples, axis=0, keepdims=True)
        centered_samples = samples - sample_mean

        def moment_bound_fn(p):
            return np.mean(np.sum(centered_samples**p, axis=1))
    for p in [1, 2]:
        Cp = moment_bound_fn(2 * p)
        results['W{}'.format(p)] = 2 * Cp**(.5 / p) * np.expm1(d2)**(.5 / p)
    return results


def divergence_bound(log_weights, *, alpha=2., log_norm_bound=None,
                     return_log_norm_bound=False):
    """Compute a bound on the alpha-divergence.

    Compute error and distance bounds between distribution `p` and `q` using
    samples from `q`.

    Parameters
    ----------
    log_weights : `array-like` of `int`, shape (n_samples,)
        log weights `log p(x_i) - log q(x_i)`, where `x_i` is sampled from `q`
        and `p` may be an unnormalized distribution.

    alpha : `float`
        order of the Renyi divergence. Must be greater than 1

    log_norm_bound : `float`
        Bound on the log normalization constant for `p` (the log marginal
        likelihood when `p` is the unnormalized log posterior).

    Returns
    -------
    dalpha : `float`
        Bound on the alpha-divergence."""
    if alpha <= 1:
        raise ValueError('alpha must be greater than 1')
    log_weights = np.asarray(log_weights)
    log_rescale = np.max(log_weights)
    rescaled_weights = np.exp(log_weights - log_rescale)**alpha
    mean_rescaled_weight = mean_and_check_mc_error(rescaled_weights,
                                                   quantity_name='CUBO')
    cubo = np.log(mean_rescaled_weight) / alpha + log_rescale
    if log_norm_bound is None:
        log_norm_bound = mean_and_check_mc_error(log_weights,
                                                 quantity_name='ELBO')
    dalpha = alpha / (alpha - 1) * (cubo - log_norm_bound)
    if return_log_norm_bound:
        return dalpha, log_norm_bound
    return dalpha


def mean_and_check_mc_error(a, atol=0.01, rtol=0.0, quantity_name=None):
    m = np.mean(a)
    s = np.std(a) / np.sqrt(a.size)
    if s > rtol * np.abs(m) + atol:  # pragma: no cover
        msg = 'significant Monte Carlo error'
        if quantity_name is not None:
            msg += ' when computing ' + quantity_name
        msg += ' (mean = {}, standard deviation = {})'.format(m, s)
        warn(msg)
    return m


_var_bound_const_1 = 2 * np.sqrt(2)
_var_bound_const_2 = 1 + 3 * np.sqrt(2)


def mean_bound(Wp):
    return Wp


def std_bound(W2):
    return W2


def var_bound(W2, var1, var2=None):
    if var2 is not None:
        min_var = np.min([var1, var2], axis=0)
    else:
        min_var = var1
    min_std = np.sqrt(min_var)
    return 2 * (min_std * W2 + W2**2)



from abc import ABC, abstractmethod

import autograd.numpy as np
import tqdm

from viabel._mc_diagnostics import MCSE, R_hat_convergence_check
from viabel._utils import Timer, StanModel_cache
from viabel.approximations import MFGaussian
from collections import defaultdict

__all__ = [
    'Optimizer',
    'StochasticGradientOptimizer',
    'RMSProp',
    'Adam',
    'Adagrad',
    'WindowedAdagrad',
    'AveragedRMSProp',
    'AveragedAdam',
    'FASO',
    'RAABBVI'
]


class Optimizer(ABC):
    """An abstract class for optimization
    """

    @abstractmethod
    def optimize(self, n_iters, objective, init_param, **kwargs):
        """
        Parameters
        ----------
        n_iters : `int`
            Number of iterations of the optimization
        objective : `function`
            Function for constructing the objective and gradient function
        init_param : `numpy.ndarray`, shape(var_param_dim,)
            Initial values of the variational parameters
        **kwargs
            Keyword arguments to pass (example: smoothed_prop)

        Returns
        ----------
        results : `dict`
            Must contain at least `opt_param`, the estimate for the optimal
            variational parameter.
        """


class StochasticGradientOptimizer(Optimizer):
    """Stochastic gradient descent.
    """

    def __init__(self, learning_rate, *, weight_decay=0, iterate_avg_prop=0.2,
                 diagnostics=False):
        """
        Parameters
        -----------
        learning_rate : `float`
            Tuning parameter that determines the step size
        weight_decay: `float`
            L2 regularization weight
        iterate_avg_prop : `float`
            Proportion of iterates to use for computing iterate average. `None`
            means no iterate averaging. The default is 0.2.
        diagnostics : `bool`, optional
            Record diagnostic information if `True`. The default is `False`.
        """
        self._learning_rate = learning_rate
        self._weight_decay = weight_decay
        if iterate_avg_prop is not None and (iterate_avg_prop > 1.0
                                             or iterate_avg_prop <= 0.0):
            raise ValueError('"iterate_avg_prop" must be None or between 0 and 1')
        self._iterate_avg_prop = iterate_avg_prop
        self._diagnostics = diagnostics
        self.reset_state()

    def reset_state(self):
        """Reset internal state of the optimizer"""
        pass

    def optimize(self, n_iters, objective, init_param, init_hamflow_model_param=None,
                 init_hamflow_rho_param=None):
        variational_param = init_param.copy()
        iap = self._iterate_avg_prop
        results = defaultdict(list)
        # value_history = []
        # variational_param_history = []
        # descent_dir_history = []
        with tqdm.trange(n_iters) as progress:
            try:
                for k in progress:
                    # take step in descent direction
                    object_val, object_grad = objective(variational_param)
                    descent_dir = self.descent_direction(object_grad)
                    variational_param = objective.update(variational_param, 
                                             self._learning_rate * descent_dir)
                    if variational_param.ndim == 2:
                        variational_param *= (1 - self._weight_decay)
                    # record state information
                    results['value_history'].append(object_val)
                    if self._diagnostics or iap is not None:
                        results['variational_param_history'].append(variational_param.copy())
                        if (iap is not None and len(results['variational_param_history']) > iap * k):
                            results['variational_param_history'].pop(0)
                    if self._diagnostics:
                        results['descent_dir_history'].append(descent_dir)
                    if k % 10 == 0:
                        avg_loss = np.mean(results['value_history'][max(0, k - 1000):k + 1])
                        progress.set_description(
                            'average loss = {:,.5g}'.format(avg_loss))
            except (KeyboardInterrupt, StopIteration):  # pragma: no cover
                # do not print log on the same line
                progress.close()
            finally:
                progress.close()
        
        if iap is not None:
            window = max(1, int(k * iap))
            results['opt_param'] = np.mean(results['variational_param_history'][-window:], axis=0)
        else:
            results['opt_param'] = variational_param.copy()
        # if descent_dir_history is not None:
        #     results['descent_dir_history'] = descent_dir_history
        results_dict = {d: np.array(h) for d, h in results.items()}
        return results_dict

    def descent_direction(self, grad):
        """Compute descent direction for optimization.

        Default implementation returns ``grad``.

        Parameters
        -----------
        grad : `numpy.ndarray`, shape(var_param_dim,)
            (stochastic) gradient of the objective function

        Returns
        ----------
        descent_dir : `numpy.ndarray`, shape(var_param_dim,)
            Descent direction of the optimization algorithm
        """
        return grad


class RMSProp(StochasticGradientOptimizer):
    """RMSProp optimization method (Hinton and Tieleman, 2012)
    
    Tracks the exponential moving average of squared gradient:
    
    .. math::
        \\nu^{(k+1)} = \\beta \\nu^{(k)} + (1-\\beta) \\hat{g}^{(k)} \\cdot  \\hat{g}^{(k)}
    
    and uses :math:`\\nu^{(k)}` to rescale the current stochastic gradient:
        
    .. math::
        \\hat{g}^{(k+1)}/\\sqrt{\\nu^{(k)}}.
    
    Parameters
    -----------
    beta : `float` optional
        Squared gradient moving average hyper parameter. The default is 0.9
    jitter: `float` optional
        Small value used for numerical stability. The default is 1e-8

    Returns
    ----------
    descent_dir : `numpy.ndarray`, shape(var_param_dim,)
        Descent direction of the optimization algorithm
        
    """

    def __init__(self, learning_rate, *, weight_decay=0, iterate_avg_prop=0.2, 
                 beta=0.9, jitter=1e-8, diagnostics=False):
        self._beta = beta
        self._jitter = jitter
        super().__init__(learning_rate, weight_decay=weight_decay,
                         iterate_avg_prop=iterate_avg_prop, 
                         diagnostics=diagnostics)

    def reset_state(self):
        """
        resetting :math:`\\nu`, the exponential moving average of squared gradient
        """
        self._avg_grad_sq = None

    def descent_direction(self, grad):
        if self._avg_grad_sq is None:
            avg_grad_sq = grad**2
        else:
            avg_grad_sq = self._avg_grad_sq
        avg_grad_sq *= self._beta
        avg_grad_sq += (1. - self._beta) * grad**2
        descent_dir = grad / np.sqrt(self._jitter + avg_grad_sq)
        self._avg_grad_sq = avg_grad_sq
        return descent_dir


class AveragedRMSProp(StochasticGradientOptimizer):
    """Averaged RMSProp optimization method (Mukkamala and Hein, 2017, §4)
    
    Uses averaged squared gradient by setting :math:`\\beta_k = 1-1/k` such that 
    
    .. math::
        \\nu^{(k+1)} = \\beta_k \\nu^{(k)} + (1-\\beta_k) \\hat{g}^{(k)} \\cdot  \\hat{g}^{(k)}.
    
    Then,
    
    .. math::
        \\nu^{(k+1)} = (k+1)^{-1} \\sum^k_{k^\\prime =0}\\hat{g}^{(k)} \\cdot  \\hat{g}^{(k)},
    
    where :math:`\\nu^{(k)}` converges to a constant almost surely under certain 
    conditions.
    
    Parameters
    -----------
    jitter: `float` optional
        Small value used for numerical stability. The default is 1e-8
    component_wise: `boolean` optional
        Indication of  component wise discent direction computation

    Returns
    ----------
    descent_dir : `numpy.ndarray`, shape(var_param_dim,)
        Descent direction of the optimization algorithm
    """
    def __init__(self, learning_rate, *, jitter=1e-8,
                 diagnostics=False, component_wise=True):
        self._jitter = jitter
        self._component_wise = component_wise
        super().__init__(learning_rate, diagnostics=diagnostics)

    def reset_state(self):
        """
        resetting :math:`\\nu` and k, the exponential moving average of squared 
        gradient and iteration respectively
        """
        self._avg_grad_sq = None
        self._t = None

    def descent_direction(self, grad):
        if self._avg_grad_sq is None:
            avg_grad_sq = grad**2
            t = 1
        else:
            avg_grad_sq = self._avg_grad_sq
            t = self._t + 1
        beta = 1 - 1/t
        avg_grad_sq *= beta
        avg_grad_sq += (1.-beta)*grad**2
        if self._component_wise:
            descent_dir = grad / np.sqrt(self._jitter+avg_grad_sq)
        else:
            descent_dir = grad / np.sqrt(self._jitter+np.sum(avg_grad_sq))
        self._avg_grad_sq = avg_grad_sq
        self._t = t
        return descent_dir
    
class Adam(StochasticGradientOptimizer):
    """Adam optimization method (Kingma and Ba, 2015)
    
    Tracks exponential moving average of the gradient as well as the 
    squared gradient: 
    
    .. math::    
        m^{(k+1)} &= \\beta_1 m^{(k)} + (1-\\beta_1) \\hat{g}^{(k)}\\\\
        \\nu^{(k+1)} &= \\beta_2 \\nu^{(k)} + (1-\\beta_2) \\hat{g}^{(k)} \\cdot \\hat{g}^{(k)}
    
    
    and uses :math:`m^{(k)}` and  :math:`\\nu^{(k)}` to rescale the current stochastic gradient:
    
    .. math::    
        m^{(k)}/\\sqrt{\\nu^{(k)}}.
        
    Parameters
    ----------
    beta1 : `float` optional
        Gradient moving average hyper parameter. The default is 0.9
    beta2 : `float` optional
        Squared gradient moving average hyper parameter. The default is 0.999
    jitter: `float` optional
        Small value used for numerical stability. The default is 1e-8
    component_wise: `boolean` optional
        Indicator for component-wise normalization of discent direction

    Returns
    ----------
    descent_dir : `numpy.ndarray`, shape(var_param_dim,)
        Descent direction of the optimization algorithm
    """
    def __init__(self, learning_rate, *, beta1=0.9, beta2=0.999, jitter=1e-8,
                 iterate_avg_prop=0.2, diagnostics=False):
        self._beta1 = beta1
        self._beta2 = beta2
        self._jitter = jitter
        super().__init__(learning_rate, iterate_avg_prop=iterate_avg_prop,
                         diagnostics=diagnostics)

    def reset_state(self):
        """
        resetting m and  :math:`\\nu`, the exponential moving average of 
        gradient and squared gradient respectively
        """
        self._momentum = None
        self._avg_grad_sq = None

    def descent_direction(self, grad):
        if self._avg_grad_sq is None:
            avg_grad_sq = grad**2
        else:
            avg_grad_sq = self._avg_grad_sq
        
        if self._momentum is None:
            momentum = grad
        else:
            momentum = self._momentum
        
        momentum *= self._beta1
        momentum += (1. - self._beta1) * grad
        avg_grad_sq *= self._beta2
        avg_grad_sq += (1. - self._beta2) * grad**2
        descent_dir = momentum / np.sqrt(self._jitter + avg_grad_sq)
        self._momentum = momentum
        self._avg_grad_sq = avg_grad_sq
        return descent_dir
    
class AveragedAdam(StochasticGradientOptimizer):
    """Averaged Adam optimization method (Mukkamala and Hein, 2017, §4)
    
    Uses averaged squared gradient by setting :math:`\\beta_k = 1-1/k` such that
    
    .. math::    
        \\nu^{(k+1)} = \\beta_k \\nu^{(k)} + (1-\\beta_k) \\hat{g}^{(k)} \\cdot \\hat{g}^{(k)}.
    Then,
    
    .. math::   
        \\nu^{(k+1)} = (k+1)^{-1} \\sum^k_{k^\\prime =0}\\hat{g}^{(k)} \\cdot  \\hat{g}^{(k)},
    where :math:`\\nu^{(k)}` converges to a constant almost surely under certain 
    conditions.
    
    Parameters
    ----------
    beta1 : `float` optional
        Gradient moving average hyper parameter. The default is 0.9
    jitter: `float` optional
        Small value used for numerical stability. The default is 1e-8
    component_wise: `boolean` optional
        Indicator for component-wise normalization of discent direction

    Returns
    ----------
    descent_dir : `numpy.ndarray`, shape(var_param_dim,)
        Descent direction of the optimization algorithm
    """
    def __init__(self, learning_rate, *, beta1=0.9, jitter=1e-8,
                 diagnostics=False, component_wise=True):
        self._beta1 = beta1
        self._jitter = jitter
        self._component_wise = component_wise
        super().__init__(learning_rate, diagnostics=diagnostics)

    def reset_state(self):
        """
        resetting m, :math:`\\nu` and, k, the exponential moving average of 
        gradient, squared gradient, and iteration respectively
        """
        self._momentum = None
        self._avg_grad_sq = None
        self._t = None

    def descent_direction(self, grad):
        if self._avg_grad_sq is None:
            avg_grad_sq = grad**2
            t = 1
        else:
            avg_grad_sq = self._avg_grad_sq
            t = self._t + 1
        if self._momentum is None:
            momentum = grad
        else:
            momentum = self._momentum
        
        momentum *= self._beta1
        momentum += (1. - self._beta1) * grad
        beta2 = 1 - 1/t
        avg_grad_sq *= beta2
        avg_grad_sq += (1. - beta2) * grad**2
        if self._component_wise:
            descent_dir = momentum / np.sqrt(self._jitter+avg_grad_sq)
        else:
            descent_dir = momentum / np.sqrt(self._jitter+np.sum(avg_grad_sq))
        self._momentum = momentum
        self._avg_grad_sq = avg_grad_sq
        self._t = t
        return descent_dir

class Adagrad(StochasticGradientOptimizer):
    """Adagrad optimization method (Duchi et al., 2011)
    
    Uses accumilated squared gradients to rescale the current stochastic
    gradient:   
        
    .. math::   
        \\frac{\\hat{g}^{(k+1)}}{\\sqrt{\\sum^k_{k^\\prime} \\hat{g}^{(k^\\prime)} \\cdot \\hat{g}^{(k^\\prime)}}}
    
    Parameters
    -----------
    jitter: `float` optional
        Small value used for numerical stability. The default is 1e-8

    Returns
    ----------
    descent_dir : `numpy.ndarray`, shape(var_param_dim,)
        Descent direction of the optimization algorithm
    """

    def __init__(self, learning_rate, *, weight_decay=0, jitter=1e-8, 
                 iterate_avg_prop=0.2, diagnostics=False):
        self._jitter = jitter
        super().__init__(learning_rate, weight_decay=weight_decay, 
                         iterate_avg_prop=iterate_avg_prop, diagnostics=diagnostics)

    def reset_state(self):
        """
        restting accumilated squared gradient
        """
        self._sum_grad_sq = 0

    def descent_direction(self, grad):
        self._sum_grad_sq += grad**2
        descent_dir = grad / np.sqrt(self._jitter + self._sum_grad_sq)
        return descent_dir

class WindowedAdagrad(StochasticGradientOptimizer):
    """Windowed Adagrad optimization method (Default optimizer in Pymc3)
    
    Uses a running window (w) to get the mean squared gradient to rescale
    the current stochastic gradient:
        
    .. math::
        \\frac{\\hat{g}^{(k+1)}}{\\sqrt{\\sum^k_{k^\\prime = k-w} \\hat{g}^{(k^\\prime)} \\cdot \\hat{g}^{(k^\\prime)}}}
    
    Parameters
    -----------
    window size : `int` optional
        Window size used to store the square of the gradients. The default is 10
    jitter: `float` optional
        Small value used for numerical stability. The default is 1e-8

    Returns
    ----------
    descent_dir : `numpy.ndarray`, shape(var_param_dim,)
        Descent direction of the optimization algorithm
    """

    def __init__(self, learning_rate, *, weight_decay=0, window_size=10, 
                 jitter=1e-8, diagnostics=False):
        self._window_size = window_size
        self._jitter = jitter
        super().__init__(learning_rate, weight_decay=weight_decay, 
                         diagnostics=diagnostics)

    def reset_state(self):
        """
        resetting the running squared gradients
        """
        self._history = []

    def descent_direction(self, grad):
        self._history.append(grad**2)
        if len(self._history) > self._window_size:
            self._history.pop(0)
        mean_grad_squared = np.mean(self._history, axis=0)
        descent_dir = grad / np.sqrt(self._jitter + mean_grad_squared)
        return descent_dir

        
class FASO(Optimizer):
    """Fixed-learning rate stochastic optimization meta-algorithm (FASO)
    
    This algorithm runs stochastic optimization with a fixed-learning rate using 
    a user specified optimization method. It determines the convergence at the 
    fixed-learning rate using the potential scale reduction factor \(\hat{R}\) and 
    combined with a Monte Carlo standard error cutoff.
    
    See more details: https://arxiv.org/pdf/2203.15945.pdf

    Parameters
    ----------
    sgo : `StochasticGradientOptimizer` object
        optimization method to use
    mcse_threshold : `float` optional
        Monte Carlo standard error threshold for convergence. The default is 0.1.
    W_min : `int`, optional
        Minimum window size for checking convergence. The default is 200.
    ESS_min : `int`, optional
        Minimum ESS for computing iterate average. Default is `W_min / 8`.
    k_check : `int`, optional
        Frequency with which to check convergence. The default is `W_min`.
    """

    def __init__(self, sgo, *, mcse_threshold=0.1, W_min=200, ESS_min=None,
                 k_check=None):
        if not isinstance(sgo, StochasticGradientOptimizer):
            raise ValueError('sgo must be a subclass of StochasticGradientOptimizer')
        self._sgo = sgo
        self._mcse_threshold = mcse_threshold
        self._W_min = W_min
        self._ESS_min = W_min // 8 if ESS_min is None else ESS_min
        self._k_check = W_min if k_check is None else k_check
        if mcse_threshold <= 0:
            raise ValueError('"mcse_threshold" must be greater than zero')
        if W_min <= 0:
            raise ValueError('"W_min" must be greater than zero')
        if self._k_check <= 0:
            raise ValueError('"k_check" must be greater than zero')
        if self._ESS_min <= 0:
            raise ValueError('"ESS_min" must be greater than zero')

    def optimize(self, n_iters, objective, init_param):
        diagnostics = self._sgo._diagnostics
        k_conv = None  # Iteration number when reached convergence
        k_stopped = None  # Iteration number when MCSE/ESS conditions met
        k_Rhat = None  # Iteration number when R hat convergence criterion met
        learning_rate = self._sgo._learning_rate
        variational_param = init_param.copy()
        history = defaultdict(list)
        iterate_average = variational_param.copy()
        if diagnostics:
            history['iterate_average_k_history'].append(0)
            history['iterate_average_history'].append(iterate_average)
        total_opt_time = 0  # total time spent on optimization
        with tqdm.trange(n_iters) as progress:
            try:
                for k in progress:
                    # take step in descent direction
                    with Timer() as opt_timer:
                        object_val, object_grad = objective(variational_param)
                        history['value_history'].append(object_val)
                        history['grad_history'].append(object_grad)
                        descent_dir = self._sgo.descent_direction(object_grad)
                        variational_param = objective.update(variational_param, learning_rate * descent_dir)
                        history['variational_param_history'].append(variational_param.copy())
                        if diagnostics:
                            history['descent_dir_history'].append(descent_dir)
                    total_opt_time += opt_timer.interval
                    # If convergence has not been reached then check for
                    # convergence using R hat
                    if k_conv is None and k % self._k_check == 0:
                        W_upper = int(0.95 * k)
                        if W_upper > self._W_min:
                            windows = np.linspace(self._W_min, W_upper, num=5, dtype=int)
                            R_hat_success, best_W = R_hat_convergence_check(
                                history['variational_param_history'], windows)
                            iterate_average = np.mean(history['variational_param_history'][-best_W:], axis=0)
                            if diagnostics:
                                history['iterate_average_k_history'].append(k)
                                history['iterate_average_history'].append(iterate_average)
                            if R_hat_success:
                                k_Rhat = k
                                k_conv = k - best_W
                                W_check = best_W  # immediately check MCSE

                    # Once convergence has been reached compute the MCSE
                    if k_conv is not None and k - k_conv == W_check:
                        W = W_check
                        converged_iterates = np.array(history['variational_param_history'][-W:])
                        iterate_average = np.mean(converged_iterates, axis=0)
                        if diagnostics and k not in history['iterate_average_k_history']:
                            history['iterate_average_k_history'].append(k)
                            history['iterate_average_history'].append(iterate_average)
                        # compute MCSE
                        with Timer() as mcse_timer:
                            if isinstance(objective.approx, MFGaussian):
                                dim = int(init_param.size/2)
                                # For MF Gaussian, use MCSE(mu/sigma,log_sigma)
                                iterate_diff = (converged_iterates[W - 2, :]
                                                - converged_iterates[W - 1, :])
                                iterate_diff_zero = iterate_diff == 0
                                # ignore constant variational parameters
                                if np.any(iterate_diff_zero):
                                    indices = np.argwhere(iterate_diff_zero)
                                    converged_iterates = np.delete(converged_iterates, indices, 1)
                                converged_log_sdevs = converged_iterates[:, -dim:]
                                mean_log_stdev = np.mean(converged_log_sdevs, axis=0)
                                ess, mcse = MCSE(converged_iterates)
                                mcse_mean = mcse[:dim] / np.exp(mean_log_stdev)
                                mcse_stdev = mcse[-dim:]
                                mcse = np.concatenate((mcse_mean, mcse_stdev))
                            else:
                                ess, mcse = MCSE(converged_iterates)
                        if diagnostics:
                            history['ess_and_mcse_k_history'].append(k)
                            history['ess_history'].append(ess)
                            history['mcse_history'].append(mcse)
                        if (np.max(mcse) < self._mcse_threshold and np.min(ess) > self._ESS_min):
                            k_stopped = k
                            break
                        else:
                            relative_mcse_time = mcse_timer.interval / W
                            relative_opt_time = total_opt_time / k
                            relative_time_ratio = relative_opt_time / relative_mcse_time
                            recheck_scale = max(1.05, 1 + 1 / np.sqrt(1 + relative_time_ratio))
                            W_check = int(recheck_scale * W_check + 1)
                    if k % self._k_check == 0:
                        avg_loss = np.mean(history['value_history'][max(0, k - 1000):k + 1])
                        R_conv = 'converged' if k_conv is not None else 'not converged'
                        progress.set_description(
                            'average loss = {:,.5g} | R hat {}|'.format(avg_loss, R_conv))
            except (KeyboardInterrupt, StopIteration):  # pragma: no cover
                # do not print log on the same line
                progress.close()
            finally:
                progress.close()
        if k_stopped is None:
            if k_conv is None:
                print('WARNING: stationarity not reached after maximum number of iterations')
                print('WARNING: try incresing the learning rate or the maximum number of '
                      'iterations')
            else:
                print('WARNING: stationarity reached but MCSE too large and/or ESS too small')
                print('WARNING: maximum MCSE = {:.3g}'.format(np.max(mcse)))
                print('WARNING: minimum ESS = {:.1f}'.format(np.min(ess)))
                # print(ess)
        else:
            print('Convergence reached at iteration', k_stopped)
        results = {d: np.array(h) for d, h in history.items()}
        results['k_conv'] = k_conv
        results['k_Rhat'] = k_Rhat
        results['k_stopped'] = k_stopped
        results['opt_param'] = iterate_average
        return results

class RAABBVI(FASO):
    """A robust, automated, and accurate BBVI optimizer (RAABBVI)
    
    This algorithm combines the FASO algorithm with a termination rule to determine
    the appropriate point where the algorithm could terminate. The termination rule is
    based on the trade-off between improved accuracy of the variational approximation
    if the current learning rate is reduced by an adaptation factor :math:`\\rho \in (0,1)`  and 
    the time required to reach that improved accuracy. If the improved accuracy
    level is large compared to the runtime then this algorithm adaptively decrease
    the learning rate and if not algorithm will be terminated.
    
    See more details: https://arxiv.org/pdf/2203.15945.pdf

    Parameters
    ----------
    sgo : `StochasticGradientOptimizer` object
        Optimizer to use for computing descent direction.
    rho : `float`, optional
        Learning rate reducing factor. The default is 0.5
    iters0 : `int`, optional
        Small iteration number. The default is 1000
        for decreased learning rate.
    accuracy_threshold : `float`, optional
        Absolute SKL accuracy threshold 
    inefficiency_threshold : `float`, optional
        Threshold for the inefficiency index.
    init_rmpsprop : `Boolean`, optional
        Indicate whether to run using RMSProp optimization method for the initial learning rate
    **kwargs:
        Keyword arguments for `FASO`
    """
    def __init__(self, sgo, *, rho=0.5, iters0=1000, accuracy_threshold=0.1, inefficiency_threshold=1.0,
                 init_rmsprop=False, **kwargs):
        super().__init__(sgo, **kwargs)
        self._iters0 = iters0
        self._rho = rho
        self._accuracy_threshold = accuracy_threshold
        self._inefficiency_threshold = inefficiency_threshold
        self._init_rmsprop = init_rmsprop
        if rho < 0 or rho > 1:
            raise ValueError('"rho" must be between zero and one')

    def weighted_linear_regression(self, model, y, x, s=9, a=0.25, n_chains=4):
        """
        weighted regression with likelihood term having the weight
        Parameters
        ----------
        model : `pystan model`
            Pystan model to conduct the sampling
        y : `numpy_ndarray`
            Response variable
        x : `numpy_ndarray`
            Predictor variable
        s : `int`, optional
            Observation weight parameter. The default is 9.
        a : `int`, optional
            Power of the weight parameter. The default is 1/4.
        n_chains: `int`, optional
            Number of sampling chains. The default is 4.
            
        Returns
        -------
        kappa : `float`
            Power
        c : `float`
            Constant
        fit : `Pystan object`
            Pystan fit object if results = True
        """
        #defining initialization function
        def initfun(log_c, sigma, kappa=None, chain_id=1):
            if kappa is None:
                return dict(log_c=log_c, sigma=sigma)
            else:
                return dict(kappa=kappa, log_c=log_c, sigma=sigma)
        N = len(y)
        w = np.array(1/(1 + np.arange(N)[::-1]**2/s)**a) #weights
        data = dict(N=N, y=y, x=x, rho=self._rho, w=w) #data
        if isinstance(self._sgo, AveragedRMSProp) or isinstance(self._sgo, AveragedAdam):
              init = [initfun(100, 5, chain_id=i) for i in range(n_chains) ] #initial values
        else:
            init = [initfun(100, 5, 0.8, chain_id=i) for i in range(n_chains) ] #initial values
        fit = model.sampling(data=data, init=init, iter=1000, chains=n_chains,
                             control=dict(adapt_delta=0.98)) #sampling from the model
        if isinstance(self._sgo, AveragedRMSProp) or isinstance(self._sgo, AveragedAdam):
            kappa = 1
        else:
            kappa = np.mean(fit['kappa'])
        log_c = np.mean(fit['log_c'])
        c = np.exp(log_c)
        return fit, kappa, c
        
    
    def wls(self, x, y, s=9, a=0.25):
        """
        weighted least squares

        Parameters
        ----------
        x : `numpy_ndarray`
            Predictor variable
        y : `numpy_ndarray`
            Response variable
        s : `int`, optional
            Observation weight parameter. The default is 9.
        a : `int`, optional
            Power of the weight parameter. The default is 1/4.

        Returns
        -------
        b0 : `float`
            Intercept
        b1 : `float`
            Slope
        """
        n = y.size
        x = np.column_stack((np.ones(n),x))
        w = np.diag(1/(1 + np.arange(n)[::-1]**2/s**2)**a) #weights
        y = np.reshape(y,(n,1))
        beta = np.linalg.inv(x.T @ w @ x) @ (x.T @ w @ y)
        return beta[0], beta[1]
        
    def convg_iteration_trend_detection(self, slope):
        """
        Detecting the relationship trend between learning
        rate and number of iterations to reach convergence

        Parameters
        ----------
        slope : `float`
            slope of the fitted regression model

        Returns
        -------
        bool
            Indicating having negative relationship or not

        """
        if slope < 0:
            return True
        else:
            return False
        

    def optimize(self, K_max, objective, init_param):
        """
        Parameters
        ----------
        K_max : `int`
            Number of iterations of the optimization
        objective: `function`
            Function for constructing the objective and gradient function
        init_param : `numpy.ndarray`, shape(var_param_dim,)
            Initial values of the variational parameters

        """
        if not objective.approx.supports_kl:
            print('WARNING: approximation family does not support KL. Using FASO.',
                  flush=True)
            return super().optimize(K_max, objective, init_param)
        k_new = -1 #Number of iterations at the given learning rate
        k = 0 #Number of times learning rate decreases
        k_total = 0 #Total number of iterations
        k_add = 0 #Iteration number when convergence reached for a fixed step size
        k_stopped_final = None #Iteration number when stopping rule criteria met
        sgo = self._sgo
        diagnostics = self._sgo._diagnostics
        if isinstance(self._sgo, AveragedRMSProp) or isinstance(self._sgo, AveragedAdam):
            reg_model = StanModel_cache(model_name='weighted_lin_regression_sgd')
        else:
            reg_model = StanModel_cache(model_name='weighted_lin_regression')
        iterate_average_curr = init_param.copy()
        history = defaultdict(list)
        history['iterate_average_curr_hist'].append(iterate_average_curr)
        history['k_mcse'].append(0)
        stopped = False
        try:
            while not stopped:
                K_max -= (k_new + 1)
                iterate_average_prev = iterate_average_curr
                if k == 0 and self._init_rmsprop: #Using RMSProp optimization initially if specified
                    rmsprop = RMSProp(learning_rate=sgo._learning_rate, diagnostics=diagnostics)
                    faso = FASO(sgo = rmsprop)
                    opt = faso.optimize(K_max, objective, iterate_average_curr)
                else:
                    opt = super().optimize(K_max, objective, iterate_average_curr)
                if opt['k_stopped'] is not None and k != 0: #removing the number of convergence iterations of initial learning rate
                    convg_iters = opt['k_stopped']
                    history['conv_iters_hist'].append(convg_iters)
                    # CI.append(convg_iters)
                iterate_average_curr = opt['opt_param']
                history['iterate_average_curr_hist'].append(iterate_average_curr)
                k_new = opt['k_stopped']
                
                # checking whether R hat convergence criteria reached and the FASO stopping criteria reached to add new number of iterations
                history['k_Rhat'].append(opt['k_Rhat'] + k_add if opt['k_Rhat'] is not None and k_new is not None else opt['k_Rhat']) 
                #checking whether R hat convergence criteria and the FASO stopping criteria reached to add new number of iterations
                history['k_conv'].append(opt['k_conv'] + k_add if opt['k_conv'] is not None and k_new is not None else opt['k_conv'])
                history['k_mcse'].append(k_new + k_add if k_new is not None else k_new)   
                history['variational_param_history'].extend(opt['variational_param_history'])
                history['value_history'].extend(opt['value_history'])
                history['grad_history'].extend(opt['grad_history'])
                
                #if user require diagnostics
                if diagnostics: 
                    history['descent_dir_history'].extend(opt['descent_dir_history'])
                    #checking convergence detection to store computed MCSE and ESS
                    if opt['k_conv'] is not None: 
                        history['ess_history'].extend(opt['ess_history'])
                        history['mcse_history'].extend(opt['mcse_history'])
                        if len(history['mcse_history']) > 0:
                            history['final_mcse_history'].append(history['mcse_history'][-1])
                        else:
                             history['final_mcse_history'].append(history['mcse_history'])
                    #saving the iteration histories at the initial learning rate
                    if k == 0: 
                        history['iterate_average_k_history'].extend(opt['iterate_average_k_history'])
                        history['iterate_average_history'].extend(opt['iterate_average_history'])
                    else: #adding previous number of iterations to the current iterations 
                        history['iterate_average_k_history'].extend(opt['iterate_average_k_history'][1:] + k_add)
                        history['iterate_average_history'].extend(opt['iterate_average_history'][1:,:])
               
                k_add = history['iterate_average_k_history'][-1]
                    
                if k_new is None:  # maximum number of iterations reached
                    break
                else: #compute the stopping criteria
                    k_total += k_new
                    sgo._learning_rate *= self._rho
                    self._mcse_threshold *= self._rho
                    if isinstance(self._sgo, AveragedRMSProp) or isinstance(self._sgo, AveragedAdam):
                        self._sgo.reset_state()
                    if len(history['learning_rate_hist']) > 0:
                        SKL = (objective.approx.kl(iterate_average_prev, iterate_average_curr) +
                               objective.approx.kl(iterate_average_curr, iterate_average_prev))
                        history['SKL_history'].append(SKL)
                        
                        # Conduct weighted linear regression to estimate parameters
                        # of SKL hat
                        if len(history['SKL_history']) > 0:
                            y_wlr = np.log(history['SKL_history'])
                            x_wlr = np.log(history['learning_rate_hist'])
                            fit, kappa, c = self.weighted_linear_regression(reg_model, y_wlr, x_wlr)
                            if diagnostics:
                                history['c_sample_hist'].append(np.exp(fit['log_c']))
                                if isinstance(self._sgo, AveragedRMSProp) or \
                                    isinstance(self._sgo, AveragedAdam):
                                    history['kappa_sample_hist'] = None
                                else:
                                    history['kappa_sample_hist'].append(fit['kappa'])
                            history['kappa_hist'].append(kappa)
                            history['c_hist'].append(c)
                            #computing the termination rule criteria 
                            if len(history['learning_rate_hist']) > 1:
                                relative_skl = (self._rho)**kappa + \
                                (self._accuracy_threshold/(np.sqrt(c) *
                                       history['learning_rate_hist'][-1]**kappa))
                                curr_iters = history['conv_iters_hist'][-1]
                                _, slope = self.wls(np.log(history['learning_rate_hist']),
                                                    np.log(history['conv_iters_hist']))
                                trend_check = self.convg_iteration_trend_detection(slope)
                                if trend_check: #if negative relationship use all observations
                                    y_wls = history['conv_iters_hist']
                                    x_wls = history['learning_rate_hist']   
                                else: #remove the initial observation
                                    y_wls = history['conv_iters_hist'][1:]
                                    x_wls = history['learning_rate_hist'][1:]
                                b0, b1 = self.wls(np.log(x_wls), np.log(y_wls))
                                pred_iters = int(np.exp(b0) * \
                                        (self._rho * history['learning_rate_hist'][-1])**b1)
                                history['predicted_iters_hist'].append(pred_iters)
                                relative_iters = pred_iters/(curr_iters + self._iters0)
                                history['stopping_crt'].append(relative_skl * relative_iters)
                                #checking whether termination rule condition satisfied
                                if relative_skl * relative_iters > self._inefficiency_threshold:
                                    stopped = True
                                    k_stopped_final = k_total
                                    history['k_stopped_final_hist'].append(k_total)
                                    break 
                                
                    history['learning_rate_hist'].append(sgo._learning_rate)
                    # LR.append(sgo._learning_rate)
                k += 1
        except (KeyboardInterrupt, StopIteration) as e:  # pragma: no cover
            pass
        if stopped:
            print('Termination rule reached at iteration', k_total)
            print('Inefficiency Index:', relative_skl * relative_iters)
        else:
            print('WARNING: maximum number of iterations reached before '
                  'stopping rule was triggered')
        results = {d: np.array(h) for d, h in history.items() if d!='k_Rhat' and d!='k_mcse' and d!='k_conv' }
        results['opt_param'] = iterate_average_curr
        results['k_stopped_final'] = k_stopped_final
        results['k_Rhat'] = history['k_Rhat']; results['k_mcse'] = history['k_mcse']
        results['k_conv'] = history['k_conv']
        return results
        


from abc import ABC, abstractmethod

import autograd.numpy as np
import autograd.numpy.random as npr
import autograd.scipy.stats.norm as norm
import autograd.scipy.stats.t as t_dist
from autograd import elementwise_grad
from autograd.scipy.linalg import sqrtm
from paragami import (
    FlattenFunctionInput, NumericArrayPattern, NumericVectorPattern, PatternDict,
    PSDSymmetricMatrixPattern)

from ._distributions import multivariate_t_logpdf

__all__ = [
    'ApproximationFamily',
    'MFGaussian',
    'MFStudentT',
    'MultivariateT',
    'NeuralNet',
    'NVPFlow'
]


class ApproximationFamily(ABC):
    """An abstract class for an variational approximation family.

    See derived classes for examples.
    """

    def __init__(self, dim, var_param_dim, supports_entropy, supports_kl):
        """
        Parameters
        ----------
        dim : `int`
            The dimension of the space the distributions in the approximation family are
            defined on.
        var_param_dim : `int`
            The dimension of the variational parameter.
        supports_entropy : `bool`
            Whether the approximation family supports closed-form entropy computation.
        supports_kl : `bool`
            Whether the approximation family supports closed-form KL divergence
            computation.
        """
        self._dim = dim
        self._var_param_dim = var_param_dim
        self._supports_entropy = supports_entropy
        self._supports_kl = supports_kl

    def init_param(self):
        """A variational parameter to use for initialization.

        Returns
        -------
        var_param : `numpy.ndarray`, shape (var_param_dim,)
        """
        return np.zeros(self.var_param_dim)

    @abstractmethod
    def sample(self, var_param, n_samples, seed=None):
        """Generate samples from the variational distribution

        Parameters
        ----------
        var_param : `numpy.ndarray`, shape (var_param_dim,)
            The variational parameter.
        n_samples : `int`
            The number of samples to generate.

        Returns
        -------
        samples : `numpy.ndarray`, shape (n_samples, var_param_dim)
        """

    def entropy(self, var_param):
        """Compute entropy of variational distribution.

        Parameters
        ----------
        var_param : `numpy.ndarray`, shape (var_param_dim,)
            The variational parameter.

        Raises
        ------
        NotImplementedError
            If entropy computation is not supported."""
        if self.supports_entropy:
            return self._entropy(var_param)
        raise NotImplementedError()

    def _entropy(self, var_param):
        raise NotImplementedError()

    @property
    def supports_entropy(self):
        """Whether the approximation family supports closed-form entropy computation."""
        return self._supports_entropy

    def kl(self, var_param0, var_param1):
        """Compute the Kullback-Leibler (KL) divergence.

        Parameters
        ----------
        var_param0, var_param1 : `numpy.ndarray`, shape (var_param_dim,)
            The variational parameters.

        Raises
        ------
        NotImplementedError
            If KL divergence computation is not supported.
        """
        if self.supports_kl:
            return self._kl(var_param0, var_param1)
        raise NotImplementedError()

    def _kl(self, var_param):
        raise NotImplementedError()

    @property
    def supports_kl(self):
        """Whether the approximation family supports closed-form KL divergence computation."""
        return self._supports_kl

    @abstractmethod
    def log_density(self, var_param, x):
        """The log density of the variational distribution.

        Parameters
        ----------
        var_param : `numpy.ndarray`, shape (var_param_dim,)
            The variational parameter.
        x : `numpy.ndarray`, shape (dim,)
            Value at which to evaluate the density."""

    @abstractmethod
    def mean_and_cov(self, var_param):
        """The mean and covariance of the variational distribution.

        Parameters
        ----------
        var_param : `numpy.ndarray`, shape (var_param_dim,)
            The variational parameter.
        """

    def pth_moment(self, var_param, p):
        """The absolute pth moment of the variational distribution.

        The absolute pth moment is given by :math:`\\mathbb{E}[|X|^p]`.

        Parameters
        ----------
        var_param : `numpy.ndarray`, shape (var_param_dim,)
            The variational parameter.
        p : `int`

        Raises
        ------
        ValueError
            If `p` value not supported"""
        if self.supports_pth_moment(p):
            return self._pth_moment(var_param, p)
        raise ValueError('p = {} is not a supported moment'.format(p))

    @abstractmethod
    def _pth_moment(self, var_param, p):
        """Get pth moment of the approximating distribution"""

    @abstractmethod
    def supports_pth_moment(self, p):
        """Whether analytically computing the pth moment is supported"""

    @property
    def dim(self):
        """Dimension of the space the distribution is defined on"""
        return self._dim

    @property
    def var_param_dim(self):
        """Dimension of the variational parameter"""
        return self._var_param_dim


def _get_mu_log_sigma_pattern(dim):
    ms_pattern = PatternDict(free_default=True)
    ms_pattern['mu'] = NumericVectorPattern(length=dim)
    ms_pattern['log_sigma'] = NumericVectorPattern(length=dim)
    return ms_pattern


class MFGaussian(ApproximationFamily):
    """A mean-field Gaussian approximation family."""

    def __init__(self, dim, seed=1):
        """Create mean field Gaussian approximation family.

        Parameters
        ----------
        dim : `int`
            dimension of the underlying parameter space
        """
        self._rs = npr.RandomState(seed)
        self._pattern = _get_mu_log_sigma_pattern(dim)
        super().__init__(dim, self._pattern.flat_length(True), True, True)

    def init_param(self):
        init_param_dict = dict(mu=np.zeros(self.dim),
                               log_sigma=2 * np.ones(self.dim))
        return self._pattern.flatten(init_param_dict)

    def sample(self, var_param, n_samples, seed=None):
        my_rs = self._rs if seed is None else npr.RandomState(seed)
        param_dict = self._pattern.fold(var_param)
        return param_dict['mu'] + np.exp(param_dict['log_sigma']) * \
            my_rs.randn(n_samples, self.dim)

    def _entropy(self, var_param):
        param_dict = self._pattern.fold(var_param)
        return 0.5 * self.dim * (1.0 + np.log(2 * np.pi)) + np.sum(param_dict['log_sigma'])

    def _kl(self, var_param0, var_param1):
        param_dict0 = self._pattern.fold(var_param0)
        param_dict1 = self._pattern.fold(var_param1)
        mean_diff = param_dict0['mu'] - param_dict1['mu']
        log_stdev_diff = param_dict0['log_sigma'] - param_dict1['log_sigma']
        return .5 * np.sum(np.exp(2 * log_stdev_diff)
                           + mean_diff ** 2 / np.exp(2 * param_dict1['log_sigma'])
                           - 2 * log_stdev_diff - 1)

    def log_density(self, var_param, x):
        if x.ndim == 1:
            x = x[np.newaxis, :]
        param_dict = self._pattern.fold(var_param)
        return np.sum(norm.logpdf(x, param_dict['mu'], 
                                  np.exp(param_dict['log_sigma'])), axis=-1)

    def mean_and_cov(self, var_param):
        param_dict = self._pattern.fold(var_param)
        return param_dict['mu'], np.diag(np.exp(2 * param_dict['log_sigma']))

    def _pth_moment(self, var_param, p):
        param_dict = self._pattern.fold(var_param)
        vars = np.exp(2 * param_dict['log_sigma'])
        if p == 2:
            return np.sum(vars)
        else:  # p == 4
            return 2 * np.sum(vars**2) + np.sum(vars)**2

    def supports_pth_moment(self, p):
        return p in [2, 4]


class MFStudentT(ApproximationFamily):
    """A mean-field Student's t approximation family."""

    def __init__(self, dim, df, seed=1):
        if df <= 2:
            raise ValueError('df must be greater than 2')
        self._df = df
        self._rs = npr.RandomState(seed)
        self._pattern = _get_mu_log_sigma_pattern(dim)
        super().__init__(dim, self._pattern.flat_length(True), True, False)

    def init_param(self):
        init_param_dict = dict(mu=np.zeros(self.dim),
                               log_sigma=2 * np.ones(self.dim))
        return self._pattern.flatten(init_param_dict)

    def sample(self, var_param, n_samples, seed=None):
        my_rs = self._rs if seed is None else npr.RandomState(seed)
        param_dict = self._pattern.fold(var_param)
        return param_dict['mu'] + np.exp(param_dict['log_sigma']) * \
            my_rs.standard_t(self.df, size=(n_samples, self.dim))

    def entropy(self, var_param):
        # ignore terms that depend only on df
        param_dict = self._pattern.fold(var_param)
        return np.sum(param_dict['log_sigma'])

    def log_density(self, var_param, x):
        if x.ndim == 1:
            x = x[np.newaxis, :]
        param_dict = self._pattern.fold(var_param)
        return np.sum(t_dist.logpdf(x, self.df, param_dict['mu'], np.exp(
            param_dict['log_sigma'])), axis=-1)

    def mean_and_cov(self, var_param):
        param_dict = self._pattern.fold(var_param)
        df = self.df
        cov = df / (df - 2) * np.diag(np.exp(2 * param_dict['log_sigma']))
        return param_dict['mu'], cov

    def _pth_moment(self, var_param, p):
        df = self.df
        if df <= p:
            raise ValueError('df must be greater than p')
        param_dict = self._pattern.fold(var_param)
        scales = np.exp(param_dict['log_sigma'])
        c = df / (df - 2)
        if p == 2:
            return c * np.sum(scales**2)
        else:  # p == 4
            return c**2 * (2 * (df - 1) / (df - 4) * np.sum(scales**4) + np.sum(scales**2)**2)

    def supports_pth_moment(self, p):
        return p in [2, 4] and p < self.df

    @property
    def df(self):
        """Degrees of freedom."""
        return self._df


def _get_mu_sigma_pattern(dim):
    ms_pattern = PatternDict(free_default=True)
    ms_pattern['mu'] = NumericVectorPattern(length=dim)
    ms_pattern['Sigma'] = PSDSymmetricMatrixPattern(size=dim)
    return ms_pattern


class MultivariateT(ApproximationFamily):
    """A full-rank multivariate t approximation family."""

    def __init__(self, dim, df, seed=1):
        if df <= 2:
            raise ValueError('df must be greater than 2')
        self._df = df
        self._rs = npr.RandomState(seed)
        self._pattern = _get_mu_sigma_pattern(dim)
        self._log_density = FlattenFunctionInput(
            lambda param_dict, x: multivariate_t_logpdf(
                x, param_dict['mu'], param_dict['Sigma'], df),
            patterns=self._pattern, free=True, argnums=0)
        super().__init__(dim, self._pattern.flat_length(True), True, False)

    def init_param(self):
        init_param_dict = dict(mu=np.zeros(self.dim),
                               Sigma=10 * np.eye(self.dim))
        return self._pattern.flatten(init_param_dict)

    def sample(self, var_param, n_samples, seed=None):
        my_rs = self._rs if seed is None else npr.RandomState(seed)
        df = self.df
        s = np.sqrt(my_rs.chisquare(df, n_samples) / df)
        param_dict = self._pattern.fold(var_param)
        z = my_rs.randn(n_samples, self.dim)
        sqrtSigma = sqrtm(param_dict['Sigma'])
        return param_dict['mu'] + np.dot(z, sqrtSigma) / s[:, np.newaxis]

    def entropy(self, var_param):
        # ignore terms that depend only on df
        param_dict = self._pattern.fold(var_param)
        return .5 * np.log(np.linalg.det(param_dict['Sigma']))

    def log_density(self, var_param, x):
        return self._log_density(var_param, x)

    def mean_and_cov(self, var_param):
        param_dict = self._pattern.fold(var_param)
        df = self.df
        return param_dict['mu'], df / (df - 2.) * param_dict['Sigma']

    def _pth_moment(self, var_param, p):
        df = self.df
        if df <= p:
            raise ValueError('df must be greater than p')
        param_dict = self._pattern.fold(var_param)
        sq_scales = np.linalg.eigvalsh(param_dict['Sigma'])
        c = df / (df - 2)
        if p == 2:
            return c * np.sum(sq_scales)
        else:  # p == 4
            return c**2 * (2 * (df - 1) / (df - 4) * np.sum(sq_scales**2) + np.sum(sq_scales)**2)

    def supports_pth_moment(self, p):
        return p in [2, 4] and p < self.df

    @property
    def df(self):
        """Degrees of freedom."""
        return self._df


class NeuralNet(ApproximationFamily):
    def __init__(self, layers_shapes, nonlinearity=np.tanh, last=np.tanh,
                 mc_samples=10000, seed=1):
        """
        Parameters
        ----------
        layers_shapes : `list of int`
            The hidden layers dimensions.
        nonlinearity : `function`
            Non linear function to apply after each layer except the last layer.
        last : `function`
            Non linear function to apply after the last layer.
        mc_samples : `int`
            Number of samples to draw internally for computing mean and cov.
        seed : `int`
            Internal seed representation.
        """
        self._pattern = PatternDict(free_default=True)
        self.mc_samples = mc_samples
        self._layers = len(layers_shapes)
        self._nonlinearity = nonlinearity
        self._last = last
        self._rs = npr.RandomState(seed)
        self.input_dim = layers_shapes[0][0]
        for layer_id in range(len(layers_shapes)):
            self._pattern[str(layer_id)] = NumericArrayPattern(shape=layers_shapes[layer_id])
            self._pattern[str(layer_id) + "_b"] = NumericArrayPattern(
                shape=[layers_shapes[layer_id][1]])

        super().__init__(layers_shapes[-1][-1], self._pattern.flat_length(True), False, False)

    def forward(self, var_param, x):
        log_det_J = np.zeros(x.shape[0])
        derivative = elementwise_grad(self._nonlinearity)
        derivative_last = elementwise_grad(self._last)
        for layer_id in range(self._layers):
            W = var_param[str(layer_id)]
            b = var_param[str(layer_id) + "_b"]
            if layer_id + 1 == self._layers:
                x = self._last(np.dot(x, W) + b)
                log_det_J += np.log(np.abs(np.dot(derivative_last(x), W.T).sum(axis=1)))
            else:
                x = self._nonlinearity(np.dot(x, W) + b)
                log_det_J += np.log(np.abs(np.dot(derivative(x), W.T).sum(axis=1)))
        return x, log_det_J

    def sample(self, var_param, n_samples):
        z_0 = npr.multivariate_normal(mean=[0] * self.input_dim,
                                      cov=np.identity(self.input_dim),
                                      size=n_samples)
        z_k, _ = self.forward(var_param, z_0)
        return z_k

    def log_density(self, var_param, x):
        raise NotImplementedError

    def mean_and_cov(self, var_param):
        samples = self.sample(var_param, self.mc_samples)
        return np.mean(samples, axis=0), np.cov(samples.T)

    def _pth_moment(self, var_param, p):
        raise NotImplementedError

    def supports_pth_moment(self, p):
        return False


class NVPFlow(ApproximationFamily):
    def __init__(self, layers_t, layers_s, mask, prior, prior_param, dim, activation=np.tanh,
                 seed=1, mc_samples=10000):
        """
        Parameters
        ----------
        layers_t : `list of int`
            The hidden layers dimensions for the translation operator.
        layers_s : `list of int`
            The hidden layers dimensions for the scaling operator.
        mask : `mask int`
            Mask to apply to the entry of each operator.
        prior : `ApproximationFamily`
            Prior for the latent space Z.
        prior_param : `numpy array`
            Parameter vector for the prior, must follow the same format as any
            variational family.
        dim : `int`
            Input dimension.
        seed : `int`
            Random seed for reproducibility.
        mc_samples : `int`
            Number of samples to draw internally for computing mean and cov.
        """
        assert len(layers_t) == len(layers_s)
        self.prior = prior
        self.prior_param = prior_param
        self.mc_samples = mc_samples
        self._dim = dim
        self._rs = npr.RandomState(seed)
        self.mask = mask
        self._pattern = PatternDict(free_default=True)
        self.t = [NeuralNet(layers_t, nonlinearity=activation, last=lambda x: x)
                  for _ in range(len(mask))]
        self.s = [NeuralNet(layers_s, nonlinearity=activation, last=np.tanh)
                  for _ in range(len(mask))]
        for layer_id in range(len(mask)):
            self._pattern[str(layer_id) + "t"] = self.t[layer_id]._pattern
            self._pattern[str(layer_id) + "s"] = self.s[layer_id]._pattern

        super().__init__(dim, self._pattern.flat_length(True), False, False)

    def g(self, var_param, z):
        """Inverse NVP flow.

        Parameters
        ----------
        var_param : `numpy array`
            Flat array of variational parameters.
        z : `numpy array`
            Latent space sample.
        """
        x = z
        param_dict = self._pattern.fold(var_param)
        for i in range(len(self.t)):
            x_ = x * self.mask[i]
            s = self.s[i].forward(param_dict[str(i) + "s"], x_)[0] * (1 - self.mask[i])
            t = self.t[i].forward(param_dict[str(i) + "t"], x_)[0] * (1 - self.mask[i])
            x = x_ + (1 - self.mask[i]) * (x * np.exp(s) + t)
        return x

    def f(self, var_param, x):
        """Forward NVP flow.

        Parameters
        ----------
        var_param : `numpy array`
            Flat array of variational parameters.
        x : `numpy array`
            Original space data.
        """
        param_dict = self._pattern.fold(var_param)
        log_det_J, z = np.zeros(x.shape[0]), x
        for i in reversed(range(len(self.t))):
            z_ = self.mask[i] * z
            s = self.s[i].forward(param_dict[str(i) + "s"], z_)[0] * (1 - self.mask[i])
            t = self.t[i].forward(param_dict[str(i) + "t"], z_)[0] * (1 - self.mask[i])
            z = (1 - self.mask[i]) * (z - t) * np.exp(-s) + z_
            log_det_J -= s.sum(axis=1)
        return z, log_det_J

    def log_density(self, var_param, x):
        z, logp = self.f(var_param, x)
        return self.prior.log_density(self.prior_param, z) + logp

    def sample(self, var_param, n_samples, seed=None):
        z_0 = self.prior.sample(self.prior_param, int(n_samples), seed=seed)
        z_k = self.g(var_param, z_0)
        return z_k

    def mean_and_cov(self, var_param):
        samples = self.sample(var_param, self.mc_samples)
        return np.mean(samples, axis=0), np.cov(samples.T)

    def _pth_moment(self, var_param, p):
        raise NotImplementedError

    def supports_pth_moment(self, p):
        return False




import os
import pickle
import shutil
import time
from hashlib import md5

import autograd.numpy as np
import pystan


def vectorize_if_needed(f, a, axis=-1):
    if a.ndim > 1:
        return np.apply_along_axis(f, axis, a)
    else:
        return f(a)


def ensure_2d(a):
    if a.ndim == 0:
        return a
    while a.ndim < 2:
        a = a[:, np.newaxis]
    return a


class Timer:
    def __init__(self):
        pass

    def __enter__(self):
        self.start = time.perf_counter()
        return self

    def __exit__(self, *args):
        self.end = time.perf_counter()
        self.interval = self.end - self.start


def _data_file_path(filename):
    """Returns the path to an internal file"""
    return os.path.abspath(os.path.join(__file__, '../stan_models', filename))


def _stan_model_cache_dir():
    return _data_file_path('cached-stan-models')


def clear_stan_model_cache():
    stan_model_dir = _stan_model_cache_dir()
    if os.path.exists(stan_model_dir):
        shutil.rmtree(stan_model_dir)


def StanModel_cache(model_code=None, model_name=None, **kwargs):
    """Use just as you would `StanModel`"""
    if model_code is None:
        if model_name is None:
            raise ValueError('Either model_code or model_name must be provided')
        model_file = _data_file_path(model_name + '.stan')
        if not os.path.isfile(model_file):
            raise ValueError('invalid model "{}"'.format(model_name))
        with open(model_file) as f:
            model_code = f.read()
    stan_model_dir = _stan_model_cache_dir()
    os.makedirs(stan_model_dir, exist_ok=True)
    code_hash = md5(model_code.encode('ascii')).hexdigest()
    if model_name is None:
        cache_fn = 'cached-model-{}.pck'.format(code_hash)
    else:
        cache_fn = 'cached-{}-{}.pck'.format(model_name, code_hash)
    cache_file = os.path.join(stan_model_dir, cache_fn)
    if os.path.exists(cache_file):
        print('Using cached StanModel{}'.format('' if model_name is None
                                                else ' for ' + model_name))
        with open(cache_file, 'rb') as f:
            sm = pickle.load(f)
    else:
        sm = pystan.StanModel(model_code=model_code, model_name=model_name)
        with open(cache_file, 'wb') as f:
            pickle.dump(sm, f)

    return sm



import codecs
import os
from setuptools import setup

PROJECT_ROOT = os.path.dirname(os.path.realpath(__file__))
REQUIREMENTS_FILE = os.path.join(PROJECT_ROOT, "requirements.txt")
REQUIREMENTS_DOCS_FILE = os.path.join(PROJECT_ROOT, "requirements-docs.txt")
REQUIREMENTS_DEV_FILE = os.path.join(PROJECT_ROOT, "requirements-dev.txt")
README_FILE = os.path.join(PROJECT_ROOT, "README.md")


def get_requirements():
    with codecs.open(REQUIREMENTS_FILE) as buff:
        return buff.read().splitlines()


def get_requirements_dev():
    with codecs.open(REQUIREMENTS_DEV_FILE) as buff:
        return buff.read().splitlines()


def get_requirements_docs():
    with codecs.open(REQUIREMENTS_DOCS_FILE) as buff:
        return buff.read().splitlines()


def get_long_description():
    with codecs.open(README_FILE, "rt") as buff:
        return buff.read()


setup(
    name = 'viabel',
    version='0.5.1',
    description='Efficient, lightweight variational inference and approximation bounds',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    author='Jonathan H. Huggins',
    author_email='huggins@bu.edu',
    url='https://github.com/jhuggins/viabel/',
    packages=['viabel'],
    include_package_data=True,
    install_requires=get_requirements(),
    extras_require={ 'docs' : get_requirements_docs(),
                     'dev' : get_requirements_dev()
                     },
    python_requires='>=3.5',
    classifiers=['Programming Language :: Python :: 3',
                 'Natural Language :: English',
                 'License :: OSI Approved :: MIT License',
                 'Intended Audience :: Science/Research',
                 'Development Status :: 2 - Pre-Alpha',
                 'Operating System :: OS Independent',
                 'Topic :: Scientific/Engineering :: Mathematics'],
    platforms='ALL',
)
